from lxml import html
import numpy as np
import bs4

with open ("leggaussquad.html","r") as myfile:
    data = myfile.read().replace('\n', '')

soup = bs4.BeautifulSoup(data,"lxml")


#for ii in xrange(2,65):

totstring = "#include \"legquadrules.h\" \n#include <string.h> \n "
for ii in xrange(2,65):
    name = "n{0}".format(ii)
    #print name
    temp = soup.find(id=name)
    tbl = temp.findAll("table")[1]
    dat = np.array([map(str,[a.string for a in row.findAll("td")]) for row in tbl.findAll("tr")][1:])[:,1:]
    #wts = [float(d) for d in dat[:,0]]
    #pts = [float(d) for d in dat[:,1]]

    wts = "static const double w{0}[{0}] = ".format(ii) +  "{" + ",".join(dat[:,0]) + "};"
    pts = "static const double pts{0}[{0}] = ".format(ii) +  "{" + ",".join(dat[:,1]) + "};"

    totstring = totstring + "\n" + wts + "\n" + pts + "\n"
    #print wts
    #print pts
    #print "\n"


totstring = totstring + "\n \n \n" 

func = "void getPtsWts(size_t n, double * pts, double * wts)\n {  \n"
func = func + "\t switch (n){\n"
for ii in xrange(2,65):
    func = func + "\t\t case {0}:\n \t\t\t memmove(pts,pts{0},{0}*sizeof(double)); \n \t\t\t memmove(wts,w{0},{0}*sizeof(double)); \n \t\t\t break; \n".format(ii) 

func = func + "\t}"
func = func + " \n}"

#print func

totstring = totstring + func + "\n"

text_file  = open("legquadrules.c","w")
text_file.write("%s" % totstring)
text_file.close()




import c3py # import the python interface to the c3 library
import numpy as np

DIM = 5  # number of features
def func2(x, param=None):
    # print("x = ",x)
    # print("shape ", x.shape)
    if param is not None:
        # print("param = ", param)
        # print(x[:, 2])
        for ii in range(DIM):
            x[:, ii] = x[:, ii] * param[ii]
    #     print(x[:, 2])
    # print("\n\n\n\n\n")
    out = np.sin(np.sum(x, axis=1))
    return out

if __name__ == "__main__":
    
    lb = [-1]*DIM    # lower bounds of features
    ub = [1]*DIM     # upper bounds of features

    # number of parameters per univariate function
    nparam = [3]*DIM 

    ## Run a rank-adaptive regression routine to approximate the first function
    ft = c3py.FunctionTrain(DIM)
    for ii in range(DIM):
        ft.set_dim_opts(ii, "legendre", lb[ii], ub[ii],
                        nparam[ii])

    verbose = 1
    init_rank = 2
    adapt = 1
    maxrank = 10
    kickrank = 2
    roundtol = 1e-8
    maxiter = 5
    scales = [2]*DIM; scales[2] = 0.2
    ft.build_approximation(func2,scales,init_rank,verbose,adapt,
                           maxrank=maxrank, round_tol=roundtol,
                           kickrank=kickrank, maxiter=maxiter)

    ## Generate test point
    test_pt = np.random.rand(DIM)*2.0-1.0
    ft_adapt_eval = ft.eval(test_pt)
    eval2s = func2(test_pt.reshape((1,DIM)))
    print("Second function with CrossApproximation:  Fteval =",ft_adapt_eval, "Should be =",eval2s)


    print("SAVE")
    filename = "saving_test.c3"
    ft.save(filename)

    ft_loaded = c3py.FunctionTrain(DIM)
    for ii in range(DIM):
        ft_loaded.set_dim_opts(ii, "legendre", lb[ii], ub[ii],
                        nparam[ii])
    ft_loaded.load(filename)

    ftdiff = ft - ft_loaded
    norm2 = ftdiff.norm2()

    print("Norm = ", norm2)
    
    

    ft2 = ft + ft + ft + ft
    print("ft2 ranks = ", ft2.get_ranks())

    ft3 = ft2 * ft
    print("ft3 ranks = ", ft3.get_ranks())

    ft3.round(eps=1e-3)
    print("ft3 rounded ranks = ", ft3.get_ranks())


    inner_product = ft.inner(ft)
    print("Inner product = ", inner_product)

    integral = ft.integrate()
    print("Integral = ", integral)




    



import c3py # import the python interface to the c3 library
import numpy as np
import matplotlib.pyplot as plt

dim = 4  # number of features
def func2(x,param=None):
    out = np.sin(np.sum(x*param,axis=1))
    return out

def gen_results(alpha):
    lb = -1      # lower bounds of features
    ub = 1       # upper bounds of features
    nparam = 3    # number of parameters per univariate function


    ## Run a rank-adaptive regression routine to approximate the first function
    ft = c3py.FunctionTrain(dim)
    for ii in range(dim):
        ft.set_dim_opts(ii,"legendre",lb,ub,nparam)

    verbose=0
    init_rank=2
    adapt=1
    ft.build_approximation(func2,alpha,init_rank,verbose,adapt)

    print("Computing Sobol Indices")
    SI = c3py.SobolIndices(ft, order=2)
    print("done")
    var = SI.get_variance()
    names = []
    mains = np.zeros((dim,))
    totals = np.zeros((dim,))
    for ii in range(dim):
        mains[ii] = SI.get_main_sensitivity(ii)
        totals[ii] = SI.get_total_sensitivity(ii)
        names.append(str(ii))

    # print("totals = ", totals)
    # print("Sum totals = ", np.sum(totals))

    # print("Mains = ", mains)
    # print("Sum mains = ", np.sum(mains))

    inter = []
    for ii in range(dim):
        for jj in range(ii+1,dim):
            val = SI.get_interaction([ii, jj])
            names.append(str(ii) + str(jj))
            inter.append(val)

    inter = np.array(inter)

    # print("sum = ", np.sum(mains) + np.sum(inter))
    left_over = var - np.sum(mains) - np.sum(inter)
    # print("leftover = ", left_over)
    names.append("other")

    all_sizes = list(mains/var) + list(inter/var) + [left_over/var]
    return names, all_sizes

if __name__ == "__main__":

    fig1, ax1 = plt.subplots(1,3)
    
    coeff = np.array([1.0**(-i) for i in range(dim)])
    names, all_sizes = gen_results(coeff)
    explode = [0] * len(all_sizes)
    explode[1] = 0.1;
    ax1[0].pie(all_sizes, explode=explode, labels=names, autopct='%.1f%%')
    ax1[0].axis('equal')
    ax1[0].set_title("Equal Coefficients")


    coeff = np.array([1.1**(-i) for i in range(dim)])
    names, all_sizes = gen_results(coeff)
    explode = [0] * len(all_sizes)
    explode[1] = 0.1;
    ax1[1].pie(all_sizes, explode=explode, labels=names, autopct='%.1f%%')
    ax1[1].axis('equal')
    ax1[1].set_title("1.1^i")

    coeff = np.array([1.5**(-i) for i in range(dim)])
    names, all_sizes = gen_results(coeff)
    explode = [0] * len(all_sizes)
    explode[1] = 0.1;
    ax1[2].pie(all_sizes, explode=explode, labels=names, autopct='%.1f%%')
    ax1[2].axis('equal')
    ax1[2].set_title("1.5^i")
    
    plt.show()






import numpy as np

totstring = "#include <string.h> \n "

name = "legpolytens.dat"
dat = np.loadtxt(name,skiprows=1,dtype=str)
    
print dat
w = dat

w  = "static const double lpolycoeffs[8000000] = {" + ",".join(w) + "};"

totstring = totstring + "\n" + w
    #print wts
    #print pts
    #print "\n"


totstring = totstring + "\n \n \n" 

text_file  = open("legtens.h","w")
text_file.write("%s" % totstring)
text_file.close()





import c3py # import the python interface to the c3 library
import pycback as pcb
import numpy as np
import matplotlib.pyplot as plt

dim = 3  # number of features
def func2(x,param=None):
    if (param is not None):
        print("param = ", param)
    out = np.sin(2*np.pi*np.sum(x,axis=1))
    return out

if __name__ == "__main__":
    
    lb = -1      # lower bounds of features
    ub = 1       # upper bounds of features
    nparam = 3   # number of parameters per univariate function

    ## Run a rank-adaptive regression routine to approximate the first function
    ft = c3py.FunctionTrain(dim)
    for ii in range(dim):
        ft.set_dim_opts(ii,"legendre",lb,ub,nparam)

    verbose=0
    init_rank=2
    adapt=1
    ft.build_approximation(func2,None,init_rank,verbose,adapt)

    # collect functions
    ranks = ft.get_ranks()
    funcs = [[]]*dim
    for ii in range(dim):
        funcs[ii] = [None]*ranks[ii]
        f, axis = plt.subplots(ranks[ii],ranks[ii+1])
        for row in range(ranks[ii]):
            funcs[ii][row] = [None]*ranks[ii+1]
            for col in range(ranks[ii+1]):
                funcs[ii][row][col] = ft.get_uni_func(ii,row,col)
                x = np.linspace(-1,1,100)
                evals = funcs[ii][row][col].eval(x)
                title = "$f_{0}^{{({1},{2})}}$".format(ii,row,col)
                if axis.ndim == 1:
                    if ii == 0:
                        axis[col].plot(x,evals,label=title)
                        axis[col].legend()
                    else:
                        axis[row].plot(x,evals,label=title)
                        axis[row].legend()
                else:
                    axis[row,col].plot(x,evals,label=title)
                    axis[row,col].legend()

    plt.show()
                


    

        
    



import c3py # import the python interface to the c3 library
import numpy as np
import matplotlib.pyplot as plt

## Define two functions
def func1(x,param=None):
    return np.sum(x,axis=1)

def func1_grad(x,param=None):
    return np.ones(x.shape[1])

def func2(x,param=None):
    return np.sin(np.sum(x,axis=1))

def func2_grad(x):
    return np.cos(np.sum(x,axis=1)) 

dim = 2                                # number of features
ndata = 100                            # number of data points
x = np.random.rand(ndata,dim)*2.0-1.0  # training samples
y1 = func1(x)                          # function values 
y2 = func2(x)                          # ditto

lb = -1                                # lower bounds of features
ub = 1                                 # upper bounds of features
nparam = 2                             # number of parameters per univariate function


## Run a rank-adaptive regression routine to approximate the first function
ft = c3py.FunctionTrain(dim)
for ii in range(dim):
    ft.set_dim_opts(ii,"legendre",lb,ub,nparam)
ft.build_data_model(ndata, x, y1, alg="AIO", obj="LS", adaptrank=1,
                    kickrank=1, roundtol=1e-10, verbose=0, store_opt_info=False)


## Run a fixed-rank regression routine to approximate the second function with stochastic gradient descent
ft_sgd = c3py.FunctionTrain(dim)
ranks = [2]*(dim+1)
ranks[0] = 1
ranks[dim] = 1
ft_sgd.set_ranks(ranks)
for ii in range(dim):
    ft_sgd.set_dim_opts(ii,"legendre",lb,ub,nparam)
ft_sgd.build_data_model(ndata,x,y2,alg="AIO",obj="LS",opt_type="SGD",verbose=0)

## Run a fixed-rank regression routine to approximate the second function
ft2 = c3py.FunctionTrain(dim)
ranks = [2]*(dim+1)
ranks[0] = 1
ranks[dim] = 1
ft2.set_ranks(ranks)
for ii in range(dim):
    ft2.set_dim_opts(ii,"legendre",lb,ub,nparam)
ft2.build_data_model(ndata,x,y2,alg="AIO",obj="LS",verbose=0)

# ## Select number of parameters through cross validation
# # ftcv = c3py.FunctionTrain(dim)
# # ranks = [4]*(dim+1)
# # ranks[0] = 1
# # ranks[dim] = 1
# # ftcv.set_ranks(ranks)
# # for ii in range(dim):
# #     ftcv.set_dim_opts(ii,"legendre",lb,ub,nparam)
# # ftcv.build_data_model(ndata,x,y2,alg="AIO",obj="LS_SPARSECORE",\
# #                       cvregweight=[1e-10,1e-8,1e-6,1e-4],kfold=3,verbose=0,cvverbose=2)


ft3 = ft + ft2  # add two function-trains
ft4 = ft * ft2  # multiply to function-trains

## Run adaptive sampling scheme
ft_adapt = c3py.FunctionTrain(dim)
for ii in range(dim):
    ft_adapt.set_dim_opts(ii,"legendre",lb,ub,nparam)
verbose=0
init_rank=2
adapt=1
ft_adapt.build_approximation(func2,None,init_rank,verbose,adapt)

ft_lin_adapt = c3py.FunctionTrain(dim)
for ii in range(dim):
    ft_lin_adapt.set_dim_opts(ii,"linelm",lb,ub,80)
verbose=0
init_rank=2
adapt=1
ft_lin_adapt.build_approximation(func2,None,init_rank,verbose,adapt)



ft.save("saving.c3")
ft_load = c3py.FunctionTrain(0)
ft_load.load("saving.c3")
## Generate test point
test_pt = np.random.rand(dim)*2.0-1.0


print("\n\n\n")

print("test_pt = ", test_pt)
grad = ft.grad_eval(test_pt)
grad_should = func1_grad(test_pt.reshape((1,dim)))
print("Grad = ", grad)
print("should be = ",grad_should)

print("\n\n\n")


ft1eval = ft.eval(test_pt) # evaluate the function train
floadeval = ft_load.eval(test_pt) # evaluate the function train
ft2eval = ft2.eval(test_pt)
ft_sgd_eval = ft_sgd.eval(test_pt)
# ftcveval = ftcv.eval(test_pt) 
ft3eval = ft3.eval(test_pt)
ft4eval = ft4.eval(test_pt)
ft_adapt_eval = ft_adapt.eval(test_pt)
ft_lin_adapt_eval = ft_lin_adapt.eval(test_pt)

eval1s = func1(test_pt.reshape((1,dim)))
eval2s = func2(test_pt.reshape((1,dim)))
eval3s = eval1s + eval2s
eval4s = eval1s * eval2s


print("Fteval =",ft1eval, "Should be =",eval1s)
print("Ft_loadeval =",floadeval, "Should be =",eval1s)
print("Second function with BFGS: Fteval =",ft2eval, "Should be =",eval2s)
print("Second function with SGD:  Fteval =",ft_sgd_eval, "Should be =",eval2s)
print("Second function with CrossApproximation:  Fteval =",ft_adapt_eval, "Should be =",eval2s)
print("Second function with CrossApproximation and linear elements:  Fteval =",ft_lin_adapt_eval, "Should be =",eval2s)
# print("Second function with CV:   Fteval =",ftcveval, "Should be =",eval2s)
print("Fteval =",ft3eval, "Should be =",eval3s)
print("Fteval =",ft4eval, "Should be =",eval4s)


print("\n\n\n")
print("Now getting optimization trajectories, run interactively and then type plt.show()")

dim = 5                                # number of features
ndata = 10000                          # number of data points
x = np.random.rand(ndata,dim)*2.0-1.0  # training samples
y1 = func2(x) + np.random.randn(ndata)*0.01  # function values 

lb = -1                                # lower bounds of features
ub = 1                                 # upper bounds of features
nparam = 10                            # number of parameters per univariate function

ranks = [2]*(dim+1)
ranks[0] = 1
ranks[dim] = 1

ft = c3py.FunctionTrain(dim)
ft.set_ranks(ranks)
for ii in range(dim):
    ft.set_dim_opts(ii,"legendre",lb,ub,nparam)
opt_ft = ft.build_data_model(ndata, x, y1, alg="AIO", obj="LS", adaptrank=0, verbose=0, store_opt_info=True)

ft_sgd = c3py.FunctionTrain(dim)
ft_sgd.set_ranks(ranks)
for ii in range(dim):
    ft_sgd.set_dim_opts(ii,"legendre",lb,ub,nparam)
opt_sgd = ft_sgd.build_data_model(ndata, x, y1, alg="AIO", obj="LS",
                                  opt_type="SGD", opt_sgd_learn_rate=1e-4,
                                  adaptrank=0, verbose=0, opt_maxiter=100, store_opt_info=True)

ft_als = c3py.FunctionTrain(dim)
ft_als.set_ranks(ranks)
for ii in range(dim):
    ft_als.set_dim_opts(ii,"legendre",lb,ub,nparam)
opt_als = ft_als.build_data_model(ndata, x, y1, alg="ALS", obj="LS",
                                  adaptrank=0, verbose=0,  store_opt_info=True)


plt.figure()
plt.plot(np.log10(opt_ft), label='AIO')
plt.plot(np.log10(opt_sgd), label='SGD')
plt.plot(np.log10(opt_als), label='ALS')
# plt.semilogy(opt_als, label='ALS')
plt.legend()


# Test serialization


# clean up memory for each function train
# ft.close()
# ft2.close()
# ft_sgd.close()
# ft3.close()
# ft4.close()





import numpy as np
v1 = 0
v2 = 1
if v1:
    totstring = "#include \"legquadrules.h\" \n#include <string.h> \n "
    for ii in xrange(2,200):
        name = "legrules/quad{0}.dat".format(ii)
        dat = np.loadtxt(name,skiprows=1,dtype=str)
    
        #print dat
        w = dat[:,1]
        p = dat[:,0]
        #w = [str(ww) for ww in dat[:,1] ]
        #p = [str(pp) for pp in dat[:,0] ]

        wts = "static const double w{0}[{0}] = ".format(ii) +  "{" + ",".join(w) + "};"
        pts = "static const double pts{0}[{0}] = ".format(ii) +  "{" + ",".join(p) + "};"

        totstring = totstring + "\n" + wts + "\n" + pts + "\n"
        #print wts
        #print pts
        #print "\n"

        
        totstring = totstring + "\n \n \n" 

    func = "void getPtsWts(size_t n, double * pts, double * wts)\n {  \n"
    func = func + "\t switch (n){\n"
    for ii in xrange(2,200):
        func = func + "\t\t case {0}:\n \t\t\t memmove(pts,pts{0},{0}*sizeof(double)); \n \t\t\t memmove(wts,w{0},{0}*sizeof(double)); \n \t\t\t break; \n".format(ii) 
        
    func = func + "\t}"
    func = func + " \n}"

    #print func
    
    totstring = totstring + func + "\n"
    
    text_file  = open("legquadrules.c","w")
    text_file.write("%s" % totstring)
    text_file.close()
elif v2 == 1:
    totstring = "#include \"legquadrules.h\" \n#include <string.h> \n "
    wts = "static double w1[1] = ".format(1) +  "{ 1.0 };"
    pts = "static double pts1[1] = ".format(1) +  "{ 0.0 };"
    totstring = totstring + "\n" + wts + "\n" + pts + "\n"
    for ii in xrange(2,200):
        name = "legrules/quad{0}.dat".format(ii)
        dat = np.loadtxt(name,skiprows=1,dtype=str)
    
        #print dat
        w = dat[:,1]
        p = dat[:,0]
        #w = [str(ww) for ww in dat[:,1] ]
        #p = [str(pp) for pp in dat[:,0] ]

        wts = "static double w{0}[{0}] = ".format(ii) +  "{" + ",".join(w) + "};"
        pts = "static double pts{0}[{0}] = ".format(ii) +  "{" + ",".join(p) + "};"

        totstring = totstring + "\n" + wts + "\n" + pts + "\n"
        #print wts
        #print pts
        #print "\n"
    totstring = totstring + "\n \n \n"
        
    totstring = totstring + "\n" + "static double * legwts[199] = { \n"
    for ii in xrange(1,199):
        totstring = totstring + "w{0}, ".format(ii)
    totstring = totstring + "w199 ".format(ii)
    totstring = totstring + "}; \n \n"

    totstring = totstring + "\n" + "static double * legpts[199] = { \n"
    for ii in xrange(1,199):
        totstring = totstring + "pts{0}, ".format(ii)
    totstring = totstring + "pts199 "
    totstring = totstring + "}; \n \n"


    func = "int getLegPtsWts2(size_t n, double * pts, double * wts)\n {  \n"
    func = func + "\t if ((n < 1) || (n > 200)){\n "
    func = func + "\t \t return 1;\n"
    func = func + "\t }\n"
    func = func + "\t pts = legpts[n-1];\n"
    func = func + "\t wts = legwts[n-1];\n"
    func = func + "\t return 0;"
    func = func + " \n}\n\n"

    totstring = totstring + func
    
    text_file  = open("legquadrules2.c","w")
    text_file.write("%s" % totstring)
    text_file.close()



""" Compressed Continuous Computation in python """

from __future__ import print_function

import sys
import os

import _c3 as c3

import numpy as np
import copy
import pycback as pcb
import atexit
# import contextlib

def poly_randu(dim, ranks, maxorder):
    """ Random function with specified ranks and legendre polynomials of maximum order"""
    bds = c3.bounding_box_init_std(dim)
    ft = c3.function_train_poly_randu(c3.LEGENDRE, bds, ranks, maxorder)
    c3.bounding_box_free(bds)
    return FunctionTrain(dim, ft=ft)

class FunctionTrain(object):
    """ Function-Train Decompositions """

    ft = None
    def __init__(self, din, ft=None):
        """ Initialize a Function Train of dimension *din* """

        self.dim = din
        self.opts = []
        self.ranks = [1]
        for ii in range(din):
            self.opts.append(None)
            self.ranks.append(1)


        self.ft = ft
        # atexit.register(self.cleanup)

    def copy(self):
        """ Copy a function train """

        ft_out = FunctionTrain(self.dim)
        ft_out.ft = c3.function_train_copy(self.ft)
        ft_out.opts = copy.deepcopy(self.opts)
        return ft_out

    def save(self, filename):
        """ Save a function train to a file with name *filename* """
        c3.function_train_save(self.ft, filename)

    def load(self, filename):
        """ Load a function-train from a file with name *filename* """

        ft = c3.function_train_load(filename)
        self.dim = c3.function_train_get_dim(ft)
        self.ft = ft

    def set_dim_opts(self, dim, ftype, lb=-1, ub=1, nparam=4,
                     kernel_height_scale=1.0, kernel_width_scale=1.0,
                     kernel_adapt_center=0,
                     lin_elem_nodes=None, kernel_nodes=None,
                     maxnum=np.inf, coeff_check=2, tol=1e-10,
                     nregions=5):
        """ Set approximation options per dimension """

        if self.opts[dim] is not None:
            raise AttributeError('cannot call set_dim_opts because was already called')

        o = dict({})
        o['type'] = ftype
        o['lb'] = lb
        o['ub'] = ub
        o['nparam'] = nparam
        o['maxnum'] = maxnum
        o['coeff_check'] = coeff_check
        o['kernel_height_scale'] = kernel_height_scale
        o['kernel_width_scale'] = kernel_width_scale
        o['kernel_adapt_center'] = kernel_adapt_center
        o['tol'] = tol
        o['nregions'] = nregions
        if lin_elem_nodes is not None:
            assert lin_elem_nodes.ndim == 1
            lin_elem_nodes = np.unique(lin_elem_nodes.round(decimals=12))
            o['nparam'] = len(lin_elem_nodes)
            o['lb'] = lin_elem_nodes[0]
            o['ub'] = lin_elem_nodes[-1]
        else:
            lin_elem_nodes = np.linspace(lb, ub, nparam)
        o['lin_elem_nodes'] = lin_elem_nodes
        if kernel_nodes is not None:
            assert kernel_nodes.ndim == 1
            kernel_nodes = np.unique(kernel_nodes.round(decimals=12))
            o['nparam'] = len(kernel_nodes)
            o['lb'] = kernel_nodes[0]
            o['ub'] = kernel_nodes[-1]
        o['kernel_nodes'] = kernel_nodes
        self.opts.insert(dim, o)

    def __convert_opts_to_c3_form__(self):

        c3_ope_opts = []
        for ii in range(self.dim):
            ftype = self.opts[ii]['type']
            lb = self.opts[ii]['lb']
            ub = self.opts[ii]['ub']
            nparam = self.opts[ii]['nparam']
            maxnum = self.opts[ii]['maxnum']
            coeff_check = self.opts[ii]['coeff_check']
            tol = self.opts[ii]['tol']
            kernel_height_scale = self.opts[ii]['kernel_height_scale']
            kernel_width_scale = self.opts[ii]['kernel_width_scale']
            kernel_adapt_center = self.opts[ii]['kernel_adapt_center']
            lin_elem_nodes = self.opts[ii]['lin_elem_nodes']
            kernel_nodes = self.opts[ii]['kernel_nodes']
            nregions = self.opts[ii]['nregions']
            if ftype == "legendre":
                c3_ope_opts.append(c3.ope_opts_alloc(c3.LEGENDRE))
                c3.ope_opts_set_lb(c3_ope_opts[ii], lb)
                c3.ope_opts_set_ub(c3_ope_opts[ii], ub)
                c3.ope_opts_set_nparams(c3_ope_opts[ii], nparam)
                c3.ope_opts_set_coeffs_check(c3_ope_opts[ii], coeff_check)
                c3.ope_opts_set_start(c3_ope_opts[ii], nparam)
                if np.isinf(maxnum) is False:
                    c3.ope_opts_set_maxnum(c3_ope_opts[ii], maxnum)
                c3.ope_opts_set_tol(c3_ope_opts[ii], tol)
            elif ftype == "hermite":
                c3_ope_opts.append(c3.ope_opts_alloc(c3.HERMITE))
                c3.ope_opts_set_nparams(c3_ope_opts[ii], nparam)
                c3.ope_opts_set_coeffs_check(c3_ope_opts[ii], coeff_check)
                c3.ope_opts_set_start(c3_ope_opts[ii], nparam)
                if np.isinf(maxnum) is False:
                    c3.ope_opts_set_maxnum(c3_ope_opts[ii], maxnum)
                c3.ope_opts_set_tol(c3_ope_opts[ii], tol)
            elif ftype == "fourier":
                c3_ope_opts.append(c3.ope_opts_alloc(c3.FOURIER))
                c3.ope_opts_set_lb(c3_ope_opts[ii], lb)
                c3.ope_opts_set_ub(c3_ope_opts[ii], ub)
                c3.ope_opts_set_nparams(c3_ope_opts[ii], nparam)
                c3.ope_opts_set_coeffs_check(c3_ope_opts[ii], coeff_check)
                c3.ope_opts_set_start(c3_ope_opts[ii], nparam)
                if np.isinf(maxnum) is False:
                    c3.ope_opts_set_maxnum(c3_ope_opts[ii], maxnum)
                c3.ope_opts_set_tol(c3_ope_opts[ii], tol)
            elif ftype == "linelm":
                c3_ope_opts.append(c3.lin_elem_exp_aopts_alloc(lin_elem_nodes))
            elif ftype == "kernel":
                if kernel_adapt_center == 0:
                    if kernel_nodes is not None:
                        x = list(kernel_nodes)
                    else:
                        x = list(np.linspace(lb, ub, nparam))
                    nparam = len(x)
                    std = np.std(x)
                    # print("standard deviation = ", std,  (ub-lb)/np.sqrt(12.0))
                    width = nparam**(-0.2) / np.sqrt(12.0) * (ub-lb)  * kernel_width_scale
                    width = nparam**(-0.2) * std * kernel_width_scale
                    c3_ope_opts.append(c3.kernel_approx_opts_gauss(nparam, x,
                                                                   kernel_height_scale,
                                                                   kernel_width_scale))
                else:
                    if kernel_nodes is not None:
                        x = list(kernel_nodes)
                        n2 = len(x)
                        nparam = 2*n2
                    else:
                        # print("here!!")
                        assert nparam % 2 == 0, "number of parameters has to be even for adaptation"
                        n2 = int(nparam/2)
                        x = list(np.linspace(lb, ub, n2))
                        # print("x = ", x)
                    std = np.std(x)
                    # width = (n2)**(-0.2) / np.sqrt(12.0) * (ub-lb)  * kernel_width_scale
                    width = n2**(-0.2) * std * kernel_width_scale
                    c3_ope_opts.append(c3.kernel_approx_opts_gauss(n2, x,
                                                                   kernel_height_scale,
                                                                   kernel_width_scale))
                    c3.kernel_approx_opts_set_center_adapt(c3_ope_opts[-1], kernel_adapt_center)
            elif ftype == "piecewise":
                c3_ope_opts.append(c3.pw_poly_opts_alloc(c3.LEGENDRE, lb, ub))
                c3.pw_poly_opts_set_maxorder(c3_ope_opts[-1], nparam)
                c3.pw_poly_opts_set_coeffs_check(c3_ope_opts[-1], coeff_check)
                c3.pw_poly_opts_set_tol(c3_ope_opts[-1], tol)
                # c3.pw_poly_opts_set_minsize(c3_ope_opts[-1], ((ub-lb)/nregions)**8)
                # c3.pw_poly_opts_set_minsize(c3_ope_opts[-1], ((ub-lb)/nregions)**3)
                c3.pw_poly_opts_set_minsize(c3_ope_opts[-1], ((ub-lb)/nregions)**2)                
                c3.pw_poly_opts_set_minsize(c3_ope_opts[-1], (ub-lb))
                c3.pw_poly_opts_set_nregions(c3_ope_opts[-1], nregions)
            else:
                raise AttributeError('No options can be specified for function type '
                                     + ftype)
        return c3_ope_opts

    def _build_approx_params(self, method=c3.REGRESS):

        c3_ope_opts = self.__convert_opts_to_c3_form__()

        c3a = c3.c3approx_create(method, self.dim)
        onedopts = []
        optnodes = []
        for ii in range(self.dim):
            ope_opts = c3_ope_opts[ii]
            if self.opts[ii]['type'] == "legendre":
                onedopts.append(c3.one_approx_opts_alloc(c3.POLYNOMIAL, ope_opts))
            elif self.opts[ii]['type'] == "hermite":
                onedopts.append(c3.one_approx_opts_alloc(c3.POLYNOMIAL, ope_opts))
            elif self.opts[ii]['type'] == "fourier":
                onedopts.append(c3.one_approx_opts_alloc(c3.POLYNOMIAL, ope_opts))
            elif self.opts[ii]['type'] == "linelm":
                onedopts.append(c3.one_approx_opts_alloc(c3.LINELM, ope_opts))
            elif self.opts[ii]['type'] == "kernel":
                onedopts.append(c3.one_approx_opts_alloc(c3.KERNEL, ope_opts))
            elif self.opts[ii]['type'] == "piecewise":
                onedopts.append(c3.one_approx_opts_alloc(c3.PIECEWISE, ope_opts))
            else:
                raise AttributeError("Don't know what to do here")


            lb = self.opts[ii]['lb']
            ub = self.opts[ii]['ub']
            if self.opts[ii]['type'] == "hermite" or self.opts[ii]['type'] == "fourier":
                nn = 50
                x = c3.linspace(lb, ub, nn)
                cv = c3.c3vector_alloc(nn, x)
                optnodes.append((x,cv))
                c3.c3approx_set_opt_opts_dim(c3a, ii, optnodes[-1][1])
                
            c3.c3approx_set_approx_opts_dim(c3a, ii, onedopts[ii])

        return c3a, onedopts, c3_ope_opts, optnodes

    def _free_approx_params(self, c3a, onedopts, low_opts, optnodes):
        for ii in range(self.dim):
            # print("ii: ", ii)
            c3.one_approx_opts_free(onedopts[ii])
            # print("woop: !")
            if self.opts[ii]['type'] == "legendre":
                c3.ope_opts_free(low_opts[ii])
            elif self.opts[ii]['type'] == "hermite":
                c3.ope_opts_free(low_opts[ii])
            elif self.opts[ii]['type'] == "fourier":
                c3.ope_opts_free(low_opts[ii])
            elif self.opts[ii]['type'] == "linelm":
                c3.lin_elem_exp_aopts_free(low_opts[ii])
            elif self.opts[ii]['type'] == "kernel":
                c3.kernel_approx_opts_free(low_opts[ii])
            elif self.opts[ii]['type'] == "piecewise":
                c3.pw_poly_opts_free(low_opts[ii])
            else:
                raise AttributeError("Don't know what to do here")

        #NOT FREING optnodes[ii][0]
        for ii in range(len(optnodes)):
            # c3.free(optnodes[ii][0])
            c3.c3vector_free(optnodes[ii][1])
            
        # print("OK")
        c3.c3approx_destroy(c3a)
        # print("OKkkk")

    def _assemble_cross_args(self, verbose, init_rank, maxrank=10, cross_tol=1e-8,
                             round_tol=1e-8, kickrank=5, maxiter=5):


        start_fibers = c3.malloc_dd(self.dim)

        c3a, onedopts, low_opts, optnodes = self._build_approx_params(c3.CROSS)


        for ii in range(self.dim):
            c3.dd_row_linspace(start_fibers, ii, self.opts[ii]['lb'],
                               self.opts[ii]['ub'], init_rank);

        
            # c3.dd_row_linspace(start_fibers, ii, 0.8,
            #                    1.2, init_rank)
            
        # NEED TO FREE OPTNODES
        
        c3.c3approx_init_cross(c3a, init_rank, verbose, start_fibers)
        c3.c3approx_set_verbose(c3a, verbose)
        c3.c3approx_set_cross_tol(c3a, cross_tol)
        c3.c3approx_set_cross_maxiter(c3a, maxiter)
        c3.c3approx_set_round_tol(c3a, round_tol)
        c3.c3approx_set_adapt_maxrank_all(c3a, maxrank)
        c3.c3approx_set_adapt_kickrank(c3a, kickrank)

        c3.free_dd(self.dim, start_fibers)
        return c3a, onedopts, low_opts, optnodes

    def build_approximation(self, f, fargs, init_rank, verbose, adapt, maxrank=10, cross_tol=1e-8,
                            round_tol=1e-8, kickrank=5, maxiter=5):
        """ Build an adaptive approximation of *f* """

        fobj = pcb.alloc_cobj()
        pcb.assign(fobj, self.dim, f, fargs)
        fw = c3.fwrap_create(self.dim, "python")
        c3.fwrap_set_pyfunc(fw, fobj)
        c3a, onedopts, low_opts, optnodes = self._assemble_cross_args(verbose, init_rank,
                                                            maxrank=maxrank,
                                                            cross_tol=cross_tol,
                                                            round_tol=round_tol,
                                                            kickrank=kickrank,
                                                            maxiter=maxiter)
        # print("do cross\n");
        self.ft = c3.c3approx_do_cross(c3a, fw, adapt)

        self._free_approx_params(c3a, onedopts, low_opts, optnodes)
        c3.fwrap_destroy(fw)


    def build_data_model(self, ndata, xdata, ydata, alg="AIO", obj="LS", verbose=0,
                         opt_type="BFGS", opt_gtol=1e-10, opt_relftol=1e-10,
                         opt_absxtol=1e-30, opt_maxiter=2000, opt_sgd_learn_rate=1e-3,
                         adaptrank=0, roundtol=1e-5, maxrank=10, kickrank=2,
                         kristoffel=False, regweight=1e-7, cvnparam=None,
                         cvregweight=None, kfold=5, cvverbose=0, als_max_sweep=20,
                         cvrank=None, norm_ydata=False, store_opt_info=False):
        """
        Note that this overwrites multiopts, and the final rank might not be the same
        as self.rank

        xdata should be ndata x dim
        """

        assert isinstance(xdata, np.ndarray)
        assert ydata.ndim == 1

        optimizer = None
        if opt_type == "BFGS":
            optimizer = c3.c3opt_create(c3.BFGS)
            c3.c3opt_set_absxtol(optimizer, opt_absxtol)
            c3.c3opt_ls_set_maxiter(optimizer, 300)
            # c3.c3opt_ls_set_alpha(optimizer, 0.1)
            # c3.c3opt_ls_set_beta(optimizer, 0.5)
        elif opt_type == "SGD":
            optimizer = c3.c3opt_create(c3.SGD)
            c3.c3opt_set_sgd_nsamples(optimizer, xdata.shape[0])
            c3.c3opt_set_sgd_learn_rate(optimizer, opt_sgd_learn_rate)
            c3.c3opt_set_absxtol(optimizer, opt_absxtol)
        else:
            raise AttributeError('Optimizer:  ' + opt_type + " is unknown")

        if store_opt_info is True:
            c3.c3opt_set_storage_options(optimizer, 1, 0, 0)
            
        if verbose > 1:
            c3.c3opt_set_verbose(optimizer, 1)

        # Set optimization options
        c3.c3opt_set_gtol(optimizer, opt_gtol)
        c3.c3opt_set_relftol(optimizer, opt_relftol)
        c3.c3opt_set_maxiter(optimizer, opt_maxiter)

        c3a, onedopts, low_opts, opt_opts = self._build_approx_params(c3.REGRESS)
        multiopts = c3.c3approx_get_approx_args(c3a)

        reg = c3.ft_regress_alloc(self.dim, multiopts, self.ranks)
        if alg == "AIO" and obj == "LS":
            c3.ft_regress_set_alg_and_obj(reg, c3.AIO, c3.FTLS)
        elif alg == "AIO" and obj == "LS_SPARSECORE":
            c3.ft_regress_set_alg_and_obj(reg, c3.AIO, c3.FTLS_SPARSEL2)
            c3.ft_regress_set_regularization_weight(reg, regweight)
        elif alg == "ALS" and obj == "LS":
            c3.ft_regress_set_alg_and_obj(reg, c3.ALS, c3.FTLS)
            c3.ft_regress_set_max_als_sweep(reg, als_max_sweep)
        elif alg == "ALS" and obj == "LS_SPARSECORE":
            c3.ft_regress_set_alg_and_obj(reg, c3.ALS, c3.FTLS_SPARSEL2)
            c3.ft_regress_set_regularization_weight(reg, regweight)
            c3.ft_regress_set_max_als_sweep(reg, als_max_sweep)
        else:
            raise AttributeError('Option combination of algorithm and objective not implemented '\
                                 + alg + obj)
        if alg == 'ALS':
            c3.ft_regress_set_als_conv_tol(reg, opt_relftol)
            
        if adaptrank != 0:
            c3.ft_regress_set_adapt(reg, adaptrank)
            c3.ft_regress_set_roundtol(reg, roundtol)
            c3.ft_regress_set_maxrank(reg, maxrank)
            c3.ft_regress_set_kickrank(reg, kickrank)
            c3.ft_regress_set_kfold(reg, kfold)

        c3.ft_regress_set_verbose(reg, verbose)

        if kristoffel is True:
            c3.ft_regress_set_kristoffel(reg, 1)

        if self.ft is not None:
            c3.function_train_free(self.ft)



        cv = None
        cvgrid = None
        if (cvnparam is not None) and (cvregweight is None) and (cvrank is None):
            cvgrid = c3.cv_opt_grid_init(1)
            c3.cv_opt_grid_add_param(cvgrid, "num_param", len(cvnparam), list(cvnparam))
        elif (cvnparam is None) and (cvregweight is not None) and (cvrank is None):
            cvgrid = c3.cv_opt_grid_init(1)
            c3.cv_opt_grid_add_param(cvgrid, "reg_weight", len(cvregweight), list(cvregweight))
        elif (cvnparam is not None) and (cvregweight is not None) and (cvrank is None):
            cvgrid = c3.cv_opt_grid_init(2)
            c3.cv_opt_grid_add_param(cvgrid, "num_param", len(cvnparam), list(cvnparam))
            c3.cv_opt_grid_add_param(cvgrid, "reg_weight", len(cvregeight), list(cvnparam))
        elif (cvnparam is not None) and (cvrank is not None):
            cvgrid = c3.cv_opt_grid_init(2)
            c3.cv_opt_grid_add_param(cvgrid, "rank", len(cvrank), list(cvrank))
            c3.cv_opt_grid_add_param(cvgrid, "num_param", len(cvnparam), list(cvnparam))

        
        

        yuse = ydata
        if norm_ydata is True:
            vmin = np.min(ydata)
            vmax = np.max(ydata)
            vdiff = vmax - vmin
            assert (vmax - vmin) > 1e-14
            yuse = ydata / vdiff - vmin / vdiff

        if cvgrid is not None:
            #print("Cross validation is not working yet!\n")
            c3.cv_opt_grid_set_verbose(cvgrid, cvverbose)

            cv = c3.cross_validate_init(self.dim, xdata.flatten(order='C'), yuse, kfold, cvverbose)
            c3.cross_validate_grid_opt(cv, cvgrid, reg, optimizer)
            c3.cv_opt_grid_free(cvgrid)
            c3.cross_validate_free(cv)


        # print("Run regression")
        self.ft = c3.ft_regress_run(reg, optimizer, xdata.flatten(order='C'), yuse)
        # print("Done!")


        if norm_ydata is True: # need to unnormalize
            ft_use = self.scale_and_shift(vdiff, vmin, c3_pointer=True)
            c3.function_train_free(self.ft)
            self.ft = ft_use

        if store_opt_info is True:
            if alg == "ALS":
                nepoch = c3.ft_regress_get_nepochs(reg)
                results = np.zeros((nepoch))
                for ii in range(nepoch):
                    results[ii] = c3.ft_regress_get_stored_fvals(reg, ii)
            else:
                nepoch = c3.c3opt_get_niters(optimizer)
                results = np.zeros((nepoch))
                for ii in range(nepoch):
                    results[ii] = c3.c3opt_get_stored_function(optimizer, ii)

            # print("nepoch", nepoch)

        c3.ft_regress_free(reg)
        c3.c3opt_free(optimizer)

        # Free built approximation options
        # print("Free params")
        self._free_approx_params(c3a, onedopts, low_opts, opt_opts)
        # print("Done Free params")

        if store_opt_info is True:
            return results


    def set_ranks(self, ranks):

        if len(ranks) != self.dim+1:
            raise AttributeError("Ranks must be a list of size dim+1, \
            with the first and last elements = 1")

        if isinstance(ranks, list):
            self.ranks = copy.deepcopy(ranks)
        else:
            self.ranks = list(copy.deepcopy(ranks))

        if ranks[0] != 1:
            print("Warning: rank[0] is not specified to 1, overwriting ")
            self.ranks[0] = 1

        if ranks[self.dim] != 1:
            print("Warning: rank[0] is not specified to 1, overwriting ")
            self.ranks[self.dim] = 1

    def eval(self, pt):
        """ Evaluate a FunctionTrain """
        assert isinstance(pt, np.ndarray)
        if pt.ndim == 1:
            return c3.function_train_eval(self.ft, pt)
        else:
            assert pt.shape[1] == self.dim
            out = np.zeros((pt.shape[0]))
            for ii, p in enumerate(pt):
                out[ii] = c3.function_train_eval(self.ft, p)
            return out

    def grad_eval(self, pt):
        """ Evaluate the gradient of a FunctionTrain """

        grad_out = np.zeros((self.dim))
        c3.function_train_gradient_eval(self.ft, pt, grad_out)
        return grad_out

    def round(self, eps=1e-14):
        """ Round a FunctionTrain """

        c3a, onedopts, low_opts, optopts = self._build_approx_params()
        multiopts = c3.c3approx_get_approx_args(c3a)
        ftc = c3.function_train_round(self.ft, eps, multiopts)
        c3.function_train_free(self.ft)
        # c3.function_train_free(self.ft)
        self.ft = ftc
        self._free_approx_params(c3a, onedopts, low_opts, optopts)

    def __add__(self, other, eps=0):
        """ Add two function trains """

        out = FunctionTrain(self.dim)
        out.ft = c3.function_train_sum(self.ft, other.ft)
        out.opts = copy.deepcopy(self.opts)
        out.round(eps)
        return out

    def __sub__(self, other, eps=0):
        """ Subtract two function trains """

        # print("subtracting!")
        temp1 = c3.function_train_copy(other.ft)
        c3.function_train_scale(temp1, -1.0)

        out_ft = FunctionTrain(self.dim)
        out_ft.opts = copy.deepcopy(self.opts)
        out_ft.ft = c3.function_train_sum(self.ft, temp1)
        out_ft.round(eps)

        c3.function_train_free(temp1)
        return out_ft

    def __mul__(self, other, eps=0):
        out = FunctionTrain(self.dim)
        out.ft = c3.function_train_product(self.ft, other.ft)
        out.opts = copy.deepcopy(self.opts)
        out.round(eps)
        return out

    def integrate(self):
        return c3.function_train_integrate(self.ft)

    def inner(self, other):
        return c3.function_train_inner(self.ft, other.ft)
    
    def scale(self, a, eps=0):
        """ f <- a*f"""
        c3.function_train_scale(self.ft, a)
        return self

    def scale_and_shift(self, scale, shift, eps=0, c3_pointer=False):

        c3a, onedopts, low_opts = self._build_approx_params()
        multiopts = c3.c3approx_get_approx_args(c3a)

        ft1 = c3.function_train_copy(self.ft)
        c3.function_train_scale(ft1, scale)

        ft2 = c3.function_train_constant(shift, multiopts)

        ft_out = FunctionTrain(self.dim)
        ft_out.opts = copy.deepcopy(self.opts)
        ft_out.ft = c3.function_train_sum(ft1, ft2)

        c3.function_train_free(ft1)
        c3.function_train_free(ft2)

        c3.function_train_round(ft_out.ft, eps, multiopts)
        self._free_approx_params(c3a, onedopts, low_opts)

        if c3_pointer is True:
            ft_ret = c3.function_train_copy(ft_out.ft)
            return ft_ret
        else:
            return ft_out

    def get_ranks(self):
        dim = c3.function_train_get_dim(self.ft)
        ranks = [1]*(dim+1)
        for ii in range(dim+1):
            ranks[ii] = c3.function_train_get_rank(self.ft, ii)
        return np.array(ranks)
        
    def norm2(self):
        return c3.function_train_norm2(self.ft)

    def expectation(self):
        return c3.function_train_integrate_weighted(self.ft)
    
    def variance(self):
        mean_val = self.expectation()
        second_moment = c3.function_train_inner_weighted(self.ft, self.ft)
        return second_moment - mean_val*mean_val

    def get_uni_func(self, dim, row, col):
        return GenericFunction(c3.function_train_get_gfuni(self.ft, dim, row, col))

    def laplace(self, eps=0.0):
        ft_out = FunctionTrain(self.dim)

        c3a, onedopts, low_opts, opt_opts = self._build_approx_params(c3.REGRESS)
        multiopts = c3.c3approx_get_approx_args(c3a)


        ft_out.ft = c3.exact_laplace(self.ft, multiopts)
        ft_out.opts = copy.deepcopy(self.opts)

        self._free_approx_params(c3a, onedopts, low_opts, opt_opts)

        if eps > 0.0:
            ft_out.round(eps=eps)

        return ft_out
        
        
    def __del__(self):
        self.close()

    def close(self):
        # print("Running Cleanup ")
        if self.ft is not None:
            c3.function_train_free(self.ft)
            self.ft = None



            
class GenericFunction(object):
    """ Univariate Functions """

    gf = None
    def __init__(self, gf):
        self.gf = c3.generic_function_copy(gf)

    def eval(self, x):
        if type(x) == list:
            return [c3.generic_function_1d_eval(self.gf, xx) for xx in x]
        elif type(x) == np.ndarray:
            assert x.ndim == 1, "Only 1d arrays handled"
            return np.array([c3.generic_function_1d_eval(self.gf, xx) for xx in x])
        else:
            return c3.generic_function_1d_eval(self.gf, x)
        
    def __del__(self):
        self.close()
        
    def close(self):
        if self.gf is not None:
            c3.generic_function_free(self.gf)
            self.gf = None

class SobolIndices(object):
    """ Sobol Sensitivity Indices """

    si = None
    def __init__(self, ft, order=None):
        if order is None:
            order = ft.dim
        self.si = c3.c3_sobol_sensitivity_calculate(ft.ft, order)

    def get_total_sensitivity(self, index):
        return c3.c3_sobol_sensitivity_get_total(self.si, index)

    def get_main_sensitivity(self, index):
        return c3.c3_sobol_sensitivity_get_main(self.si, index)

    def get_variance(self):
        return c3.c3_sobol_sensitivity_get_variance(self.si)
    
    def get_interaction(self, variables):
        # assert len(variables) == 2, "Only effects between two variables is currently supported"
        for ii in range(len(variables)-1):
            assert variables[ii] < variables[ii+1], \
                "Sobol index variables must be ordered with v[i] < V[j] for i < j"
        
        ret = c3.c3_sobol_sensitivity_get_interaction(self.si, variables)
        return ret
        
    def __del__(self):
        self.close()
        
    def close(self):
        if self.si is not None:
            c3.c3_sobol_sensitivity_free(self.si)
            self.si = None



import c3py # import the python interface to the c3 library
import numpy as np
import matplotlib.pyplot as plt

def func(x):
    return x[:, 0]**2 + x[:, 1]**2 + x[:, 0] * x[:, 1] + np.sin(np.sum(x, axis=1)) \
        + x[:,0] * x[:,-1] + x[:, -1]**2.0 * np.cos(x[:, 0])


LB = -1.0                              # lower bounds of features
UB = 1.0                               # upper bounds of features
DIM = 20                               # number of features

Ntrain = 100
X = np.random.rand(Ntrain, 2)
Y = func(X)

Ntest = 1000
Xtest = np.random.rand(Ntest, 2)
Ytest = func(Xtest)


def build_ft_regress(xdata, ydata, nparam=2, init_rank=5, adaptrank=1, verbose=0):
    dim = xdata.shape[1]
    ndata = xdata.shape[0]
    
    ft = c3py.FunctionTrain(dim)
    for ii in range(dim):
        ft.set_dim_opts(ii, "legendre", LB, UB, nparam)

    ranks = [init_rank]*(dim+1)
    ranks[0] = 1
    ranks[dim] = 1
    ft.set_ranks(ranks)
    ft.build_data_model(ndata, xdata, ydata,
                        alg="AIO", obj="LS", adaptrank=adaptrank,
                        kickrank=1, roundtol=1e-10, verbose=verbose)
    return ft

print("X shape ", X.shape)
print("Y shape ", Y.shape)
ft_regress = build_ft_regress(X, Y, nparam=5, init_rank=1, adaptrank=1, verbose=0)

print("Xtest.shape ", Xtest.shape)
print("Ytest.shape ", Ytest.shape)
ft_evals = ft_regress.eval(Xtest)

err = np.linalg.norm(ft_evals - Ytest) / np.linalg.norm(Ytest)

print("Relmse = ", err)
plt.figure()
plt.plot(ft_evals, Ytest, 'o')
plt.show()




# adapted from https://stackoverflow.com/a/48015772
import sys
if sys.platform == 'darwin':
    from distutils import sysconfig
    vars = sysconfig.get_config_vars()
    vars['LDSHARED'] = vars['LDSHARED'].replace('-bundle', '-dynamiclib')
    my_link_args=["-Wl,-rpath,@loader_path"]
    my_runtime_dirs = []
else:
    #my_link_args=["-Xlinker","-Wl,-rpath,$ORIGIN"]
    my_link_args=[]#"-Xlinker","-Wl,-rpath,$ORIGIN"]
    my_runtime_dirs=['$ORIGIN/']

import os
import pathlib

from setuptools import setup, find_packages, Extension
from setuptools.command.build_ext import build_ext as build_ext_orig

import numpy

try:
    numpy_include = numpy.get_include()
except AttributeError:
    numpy_include  = numpy.get_numpy_include()

PACKAGE_NAME = "c3py"
SOURCE_DIR = pathlib.Path().absolute()

from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory,"README.md"), encoding='utf-8') as f:
    long_description = f.read()

class CMakeExtension(Extension):
    """
    An extension to run the cmake build

    This simply overrides the base extension class so that setuptools
    doesnt try to build your sources for you
    """

    def __init__(self, name, sources=[]):
        print("Add Extension ", name)
        super().__init__(name=name, sources=sources)

class build_ext(build_ext_orig):

    def run(self):
        for ext in self.extensions:
            if ext.name == '_c3_notsure':
                self.build_cmake(ext)

            if ext.name == "_c3":
                self.build_swig(ext)
                
        super().run()

    def build_swig(self, ext):
        self.announce("Building Swig", level=3)
        ext.library_dirs = [os.path.join("build", s) for s in os.listdir("build") if os.path.splitext(s)[0].startswith('lib')]
        
        print(ext.library_dirs)
        print(ext.runtime_library_dirs)
        # exit(1)
        # super().build(ext)
    
    def build_cmake(self, ext):

        ext.extra_compile_args=['-install_name @rpath/libc3.dylib']
        # print(ext.extra_compile_args)        
        # exit(1)

        print("Creating Tempory File")
        cwd = pathlib.Path().absolute()
        build_temp = pathlib.Path(self.build_temp)
        build_temp.mkdir(parents=True, exist_ok=True)
        extdir = pathlib.Path(self.get_ext_fullpath(ext.name))
        extdir.mkdir(parents=True, exist_ok=True)

        # example of cmake args
        config = 'Debug' if self.debug else 'Release'
        cmake_args = [ '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + str(extdir.parent.absolute()),
                       '-DCMAKE_BUILD_TYPE=' + config,
                       '-DBUILD_TESTS=OFF',
                       '-DBUILD_EXAMPLES=OFF',
                       '-DBUILD_PROFILING=OFF',
                       '-DBUILD_BENCHMARKS=OFF',                    
                       '-DBUILD_UTILITIES=OFF',
                       '-DMAKE_PYTHON_WRAPPERS=ON']

        # example of build args
        build_args = [
            '--config', config,
            '--', '-j4'
        ]

        print("Temporary file = ", build_temp)
        os.chdir(str(build_temp))
        self.spawn(['cmake', str(cwd)] + cmake_args)
        if not self.dry_run:
            self.spawn(['cmake', '--build', '.'] + build_args)
        os.chdir(str(cwd))

pcback = Extension('pycback',
                   sources = ['wrappers/python/callback/python_caller.c'],
                   include_dirs = [
                       numpy_include,
                   ],
                   language='c',
                   extra_compile_args=['-std=c99'])

c3py = Extension("_c3",
                 sources =["wrappers/python/c3swig/c3.i",
                 ],
                 swig_opts=["-py3",
                             "-Ic3"],
                 define_macros =[('COMPILE_WITH_PYTHON',True)],
                 undef_macros = [],
                 language='c',                 
                 runtime_library_dirs= my_runtime_dirs, 
                 # library_dirs=[''],                 
                 include_dirs=["c3/lib_clinalg", 
                               "c3/lib_interface",
                               "c3/lib_funcs",
                               "c3/lib_linalg",
                               "c3/lib_array",
                               "c3/lib_fft",
                               "c3/lib_quadrature",
                               "c3/lib_optimization",
                               "c3/lib_stringmanip",
                               "c3/lib_superlearn",
                               "c3/lib_tensor",                               
                               "c3/lib_probability",
                               numpy_include,
                 ],
                 extra_compile_args = ['-std=c99'],
                 extra_link_args =  my_link_args, 
                 libraries=["c3"]
)

setup(name='c3py',
      author="Alex Gorodetsky",
      author_email="alex@alexgorodetsky.com", 
      version='0.0.4.1',
      description="Compressed Continuous Computation Library in Python",
      long_description=long_description,
      long_description_content_type="text/markdown",
      keywords="highd-approximation machine-learning uncertainty-quantification tensors",
      url="https://github.com/goroda/compressed-continuous-computation",
      package_dir={'': 'wrappers/python/c3py'},
      py_modules=['c3py'],
      ext_modules=[CMakeExtension(name="_c3_notsure"),
                   pcback, c3py],
      cmdclass={
          'build_ext' : build_ext,
      },
      install_requires=[
          'numpy', #'cmake', 
      ],
      # license='BSD',
      classifiers=[
          "Natural Language :: English",
          "Programming Language :: C",
          "Programming Language :: Python",
          "Programming Language :: Python :: 3 :: Only",
          "License :: OSI Approved :: BSD License",
          "Topic :: Scientific/Engineering :: Mathematics"
      ]
)




import c3py # import the python interface to the c3 library
import numpy as np

## Define two functions
def func1(x,param=None):
    return np.sum(x,axis=1)

def func2(x,param=None):
    return np.sum(x**2,axis=1)

DIM = 10                                # number of features
NDATA = 100                            # number of data points
X = np.random.rand(NDATA,DIM)*2.0-1.0  # training samples
Y = func1(X)                           # function values 

LB = -1                                # lower bounds of features
UB = 1                                 # upper bounds of features
NPARAM = 2                             # number of parameters per univariate function


# This should output an equal amount of allocates and trues
def build_ft_adapt(dim):
    ft = c3py.FunctionTrain(dim)
    for ii in range(dim):
        ft.set_dim_opts(ii, "legendre", LB, UB, NPARAM)
    verbose = 0
    init_rank = 2
    adapt = 0
    ft.build_approximation(func1, None, init_rank, verbose, adapt)
    return ft

def build_ft_regress(dim):
    ft = c3py.FunctionTrain(dim)
    for ii in range(dim):
        ft.set_dim_opts(ii, "legendre", LB, UB, NPARAM)

    ranks = [5]*(DIM+1)
    ranks[0] = 1
    ranks[DIM] = 1
    ft.set_ranks(ranks)
    verbose = 0
    adaptrank = 1
    ft.build_data_model(NDATA, X, Y,
                        alg="AIO", obj="LS", adaptrank=adaptrank,
                        kickrank=1, roundtol=1e-10, verbose=verbose)
    return ft

ft_adapt = build_ft_adapt(DIM)
ft_regress = build_ft_regress(DIM)
ft_diff = ft_adapt - ft_regress

for ii in range(100):
    pt = np.random.rand(DIM)*2.0-1.0
    diff_eval = ft_diff.eval(pt)
    adapt_eval = ft_adapt.eval(pt)
    regress_eval = ft_regress.eval(pt)
    print("adapt_eval, regress_eval, diff_eval ", adapt_eval, regress_eval, diff_eval)
    # assert np.abs(diff_eval) < 1e-3

print("\n\n\n")
print("Tests Passed!")
print("Bye!")
